import matplotlib.pyplot as plt

def toplot(ts, filename = None, **kwargs):
    '''To plot formatter'''
    fig = plt.figure()
    ax = fig.add_subplot(111)
    dates = list(ts.dates())
    ax.plot(dates, ts.values())
    #ax.xaxis.set_major_locator(years)
    #ax.xaxis.set_major_formatter(yearsFmt)
    #ax.xaxis.set_minor_locator(months)
    #datemin = datetime.date(r.date.min().year, 1, 1)
    #datemax = datetime.date(r.date.max().year+1, 1, 1)
    #ax.set_xlim(datemin, datemax)
    
    # format the coords message box
    #def price(x): return '$%1.2f'%x
    #ax.format_xdata = mdates.DateFormatter('%Y-%m-%d')
    #ax.format_ydata = price
    ax.grid(True)
    
    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()
    
    ## add legend or title
    names = ts.name.split('__')
    if len(names) == 1:
        ##add title
        title = names[0]
        #fontsize = kwargs.get('title_fontsize', 7)
        fontweight = kwargs.get('title_fontweight', 'bold')
        ax.set_title(title, fontweight=fontweight)#,fontsize=fontsize,         
    else:
        ##add legend
        loc = kwargs.get('legend_location','best')
        ncol = kwargs.get('legend_ncol', 2)
        ax.legend(names, loc=loc, ncol=ncol)
    
    return plt
    


from dynts.tests import tsinterface
from dynts.tests import testop

class TestZooTS(tsinterface.TestTS):
    backend = 'zoo'

class TestNumpyTS(tsinterface.TestTS):
    backend = 'numpy'


class TestOpZooTS(testop.TestOperators):
    backend = 'zoo'

class TestOpNumpyTS(testop.TestOperators):
    backend = 'numpy'


#class TestRmetricsTS(tsbase.TestTS):
#    backend = 'rmetrics'
    


import unittest
import numpy as np
from datetime import date
from itertools import izip

from dynts import timeseries, evaluate, tsname
from dynts.utils.populate import populate, datepopulate, randomts


class TestCase(unittest.TestCase):
    backend = None
    
    def __init__(self,*args,**kwargs):
        super(TestCase,self).__init__(*args,**kwargs)
        self.np = np
        self.timeseries = timeseries
        self.evaluate = evaluate
        self.tsname = tsname
        self.datepopulate = datepopulate
        self.populate = populate
        self.randomts = randomts
        
    def getdata(self, size = 100, cols = 1, delta = 1, start = None):
        dates = self.datepopulate(size = size, delta = delta)
        data = self.populate(size = size, cols = cols)
        return dates,data
        
    def getts(self, returndata = False, delta = 1, cols = 1, size = 100):
        dates,data = self.getdata(size,cols,delta)
        ts   = self.timeseries(name = 'test', date = dates, data = data, backend = self.backend)
        self.assertEqual(ts.shape,(size,cols))
        self.assertEqual(len(ts),size)
        self.assertEqual(ts.count(),cols)
        if returndata:
            return ts,list(dates),list(data)
        else:
            return ts
        
    def isiterable(self, a):
        return hasattr(a,'__iter__')

    def assertAlmostEqual(self, a, b): #copied from pandas
        if self.isiterable(a):
            np.testing.assert_(self.isiterable(b))
            np.testing.assert_equal(len(a), len(b))
            for i in xrange(len(a)):
                self.assertAlmostEqual(a[i], b[i])
            return True

        err_msg = lambda a, b: 'expected %.5f but got %.5f' % (a, b)

        if np.isnan(a):
            np.testing.assert_(np.isnan(b))
            return

        # case for zero
        if abs(a) < 1e-5:
            np.testing.assert_almost_equal(a, b, decimal=5, err_msg=err_msg(a, b), verbose=False)
        else:
            np.testing.assert_almost_equal(1, a/b, decimal=5, err_msg=err_msg(a, b), verbose=False)
            
    def assertEqual(self, exp, recv, msg = None):
        if msg is None:
            msg = "Values do not match expected %s, received %s" %(exp, recv)
        unittest.TestCase.assertEqual(self, exp,recv,msg)

    def check_dates(self, ts, dts):
        ts_dts = list(ts.dates())
        self._check_vectors(ts_dts, dts, equal=True)
        
    def check_values(self, ts, vals):
        ts_vals = list(ts.values())
        self._check_vectors(ts_vals, vals, equal=False)

    def _check_vectors(self, v1, v2, equal=True):
        lv1 = len(v1)
        lv2 = len(v2)
        self.assertEqual(lv1, lv2, "Vectors are of different lengths %s, %s" %(lv1, lv2))

        for a,b  in izip(v1,v2):
            if equal:
                self.assertEqual(a, b)
            else:
                self.assertAlmostEqual(a,b)



import numpy as np

from dynts.conf import settings
from dynts.exceptions import ExpressionError

_ops = {'add' : lambda x,y : x+y,
        'sub' : lambda x,y : x-y,
        'mul' : lambda x,y : x*y,
        'div' : lambda x,y : x/y,
        }

def _get_op(op_name):
    global _ops
    op = _ops.get(op_name, None)
    if op is None:
        raise ExpressionError('There is no op called %s.' % op_name)
    return op


def binOp(op, indx, amap, bmap, fill_vec):
    '''
    Combines the values from two map objects using the indx values
    using the op operator. In situations where there is a missing value
    it will use the callable function handle_missing
    '''
    def op_or_missing(id):
        va = amap.get(id, None)
        vb = bmap.get(id, None)
        if va is None or vb is None:
            result = fill_vec #This should create as many elements as the number of columns!?
        else:
            try:
                result = op(va, vb)
            except Exception, e:
                result = None
            if result == None:
                result = fill_vec
            return result
    seq_arys = map(op_or_missing, indx)
    data = np.vstack(seq_arys)
    return data


def applyfn(op, v1, v2, fill_vec):
    def op_or_missing(a,b):
        try:
            result = op(a,b)
        except Exception, e:
            result = None
        if result is None:
                result = fill_vec
        return result
    if len(v1) != len(v2):
        msg = "Vectors %s, %s are different lengths %s, %s" %(v1, v2, len(v1),
                len(v2))
        raise ExpressionError(msg)
    rt = map(op_or_missing, v1, v2)
    return rt


def _toVec(shape, val):
    '''
    takes a single value and creates a vecotor / matrix with that value filled
    in it
    '''
    mat = np.empty(shape)
    mat.fill(val)
    return mat

def _create_fill_vec(ts, fill_fn):
    shape = (ts.count(),)
    fill = _toVec(shape, fill_fn())
    return fill

def _handle_scalar_ts(op_name, op, scalar, ts, fill_fn):
    fill_vec = _create_fill_vec(ts, fill_fn)
    values = ts.values()
    shape = values.shape
    v2 = _toVec(shape, scalar)
    
    dts = ts.dates()
    result = applyfn(op, v2, values, fill_vec)
    return dts, result

def _handle_ts_scalar(op_name, op, ts, scalar, fill_fn):
    fill_vec = _create_fill_vec(ts, fill_fn)
    values = ts.values()
    shape = values.shape
    v2 = _toVec(shape, scalar)
    dts = ts.dates()
    result = applyfn(op, values, v2, fill_vec)
    return dts, result

def _handle_ts_ts(op_name, op, ts, ts2, all, fill_fn):
    if ts.count() != ts2.count():
        raise ExpressionError("Cannot %s two timeseries with different number of series." % op_name)
    dts1 = set(ts.dates())
    if all:
        indx = dts1.union(ts2.dates())
    else:
        indx = dts1.intersection(ts2.dates())
    hash = ts.ashash()
    hash2 = ts2.ashash()
    
    fill = _create_fill_vec(ts, fill_fn)
    #fill = np.array([fill_fn()])
    for dt in indx:
        v  = hash.get(dt,None)
        v2 = hash2.get(dt,None)
        if v is None or v2 is None:
            v = fill
        else:
            v = op(v,v2)
        hash[dt] = v
    new_ts = hash.getts()
    rt = zip(*new_ts.items())
    return rt

def _handle_ts_or_scalar(op_name, ts1, ts2, all = True, fill = None, name = None):
    '''
    this is the main entry point for any arithmetic type function performed on a timeseries
    and/or a scalar. 
    op_name - name of the function to be performed
    ts1, ts2 - timeseries or scalars that the function is to performed over
    all - whether all dates should be included in the result
    fill - the value that should be used to represent "missing values"
    name - the name of the resulting time series
    '''
    from dynts import istimeseries
    op = _get_op(op_name)
    fill = fill if fill is not None else settings.missing_value
    if callable(fill):
        fill_fn = fill
    else:
        fill_fn = lambda : fill

    name = name or '%s(%s,%s)' % (op_name,ts1,ts2)
    ts = None
    if istimeseries(ts1):
        ts = ts1
        if istimeseries(ts2):
            dts, data =  _handle_ts_ts(op_name, op, ts1, ts2, all, fill_fn)
            
        else:
            dts, data = _handle_ts_scalar(op_name, op, ts1, ts2, fill_fn)
    else:
        if istimeseries(ts2):
            ts = ts2
            dts, data = _handle_scalar_ts(op_name, op, ts1, ts2, fill_fn)
        else:
            return op(ts1,ts2)
        
    return ts.clone(date = dts, data = data, name = name)

def ts_fn(op_name):
    fn = lambda  *args,  **kwargs : _handle_ts_or_scalar(op_name, *args, **kwargs)
    return fn

add = ts_fn('add')
sub = ts_fn('sub')
mul = ts_fn('mul')
div = ts_fn('div')



import random
from dynts.test import TestCase
from dynts.backends import ops
from dynts import exceptions 

__all__ = ['TestOperators',
           ]


class TestOperators(TestCase):

    def testTSArithOperators(self):
        ts1, dates1, data1 = self.getts(returndata = True, cols = 2)
        ts2, dates2, data2 = self.getts(returndata = True, cols = 2)
        for op in ops.values():
           # print op
            ts3 = op(ts1,ts2)
            exp = map(op, data1, data2)
            self.check_dates(ts3, dates1) #the dates should be the same
            self.check_values(ts3, exp) #the values should be 

    def testScalarArithOperators(self):
        delta = random.uniform(1.0,100.0)
        ts, dates, data = self.getts(returndata = True)
        for op in ops.values():
            new_ts = op(ts, delta)
            curry_op = lambda x : op(x, delta)

            exp = map(curry_op, data)
            self.check_dates(new_ts, dates)
            self.check_values(new_ts, exp)
    
    def testArithOpWithDiffNumberCols(self):
        '''
        you can't perform arithmetic on timeseries that each have a different number of columns
        '''
        ts1 = self.getts(cols = 2)
        ts2 = self.getts(cols = 1)
        
        op = ops.values()[0]
        
        curry_fn = lambda : op(ts1, ts2)
        self.assertRaises(exceptions.ExpressionError, curry_fn)
    
    def testArithOpWithMissingDates(self):
        ts1, dates1, data1 = self.getts(returndata = True, cols = 2)
        ts2, dates2, data2 = self.getts(returndata = True, cols = 2, delta = 2)
        
        all_dates = list(set(dates1).union(set(dates2)))
        all_dates.sort()
        op = ops.values()[0]
        ts3 = op(ts1,ts2)
        self.check_dates(ts3, all_dates) ##Check that all the dates are in the return series by default
        
        btree = ts3.asbtree()
        not_in_dates2 = set(dates1) - set(dates2)
        self._check_missing_dts(not_in_dates2, btree)
        
        not_in_dates1 = set(dates2) - set(dates1)
        self._check_missing_dts(not_in_dates1, btree)
        
    def _check_missing_dts(self, dts, btree):
        from dynts.conf import settings
        ismissing = settings.ismissing
        
        for dt in  dts:
            vals = btree[dt]
            for val in vals:
                self.assertTrue(ismissing(val))
        


'''General timeseries function'''


def better_ts_function(f):
    '''Decorator which check if timeseries has a better implementation of the function'''
    fname = f.__name__
    
    def _(ts, *args, **kwargs):
        func = getattr(ts,fname,None)
        if func:
            return func(*args, **kwargs)
        else:
            return f(ts, *args, **kwargs)
    
    _.__name__ = fname
    
    return _ 


@better_ts_function
def zscore(ts, **kwargs):
    '''Rolling Z-Score statistics. The Z-score is more formally known as
``standardised residuals``. To calculate the standardised residuals of a data set,
the average value and the standard deviation of the data value have to be estimated.

.. math::

    z = \frac{x - \mu(x)}{\sigma(x)}
'''
    m = ts.rollmean(**kwargs)
    s = ts.rollstddev(**kwargs)
    result = (ts - m)/s
    name = kwargs.get('name', None)
    if name:
        result.name = name
    return result

@better_ts_function
def prange(ts, **kwargs):
    '''Rolling Percentage range. Value between 0 and 1 indicating the position in the rolling range.'''
    mi = ts.rollmin(**kwargs)
    ma = ts.rollmax(**kwargs)
    return (ts - mi)/(ma - mi)



import unittest
import dynts
import itertools

__all__ = ['FunctionTest',
           ]
class FunctionTest(unittest.TestCase):
    '''
    This is a dynamic unittest function generating class
    The idea is that all functions should be tested against all the backends
    The list of functions is stored as a constant in the dsl module, while the 
    
    '''
    
    _initialised = False
    
    
    @classmethod
    def initialise(cls, backends, functions):
        initiated = cls._initialised
        if initiated:
            return None
        else:
            for backend, function in itertools.product(backends, functions):
                cls.create_test(function, backend)
        
        cls._initialised = True
            
    
    @classmethod    
    def create_test(cls, function, backend):
        fn = lambda self: self.allchecks(function, backend)
        test_name = 'test_%s_%s' %(function, backend)
        setattr(cls, test_name, fn)
            
        
    def allchecks(self, function, backend):
        self.checkNoParameters(function, backend)
        self.checkTwoSeries(function, backend)
        self.checkWindowParameter(function, backend)
        
    def assertTimeSeries(self, data, msg = None):
        if msg is None:
            msg = "Expected a timeseries received %s instead" %data.__class__
        self.assertTrue(dynts.istimeseries(data), msg)
    
    def checkNoParameters(self, function, backend):
        '''Test function with zero parameters against a given backend'''
        expression = '%s(GOOG)' % function
        result = dynts.evaluate(expression, backend = backend)
        self.assertEqual(str(result),expression)
        self.assertEqual(len(result.data),1)
        data = result.ts()
        self.assertTimeSeries(data)
        self.assertEqual(data.count(),1)
        
    def checkWindowParameter(self, function, backend):
        '''Test mean function with one parameter'''
        e = '%(f)s(GOOG,window=30),%(f)s(GOOG,window=60)' % {'f':function}
        result = dynts.evaluate(e, backend = backend)
        symbols = result.expression.symbols()
        self.assertEqual(len(symbols),1)
        self.assertEqual(len(result.data),1)
        data = result.ts()
        self.assertTimeSeries(data)
        self.assertEqual(data.count(),2)
        
    def checkTwoSeries(self, function, backend):
        e = '%s(GOOG,YHOO)' % function
        result = dynts.evaluate(e, backend = backend)
        symbols = result.expression.symbols()
        self.assertEqual(len(symbols),2)
        self.assertEqual(len(result.data),2)
        data = result.ts()
        self.assertTimeSeries(data)
        self.assertEqual(data.count(),2)
        
FunctionTest.initialise(dynts.backends.BACKENDS.keys(),
                        functions = dynts.dsl.functions.FUNCTIONS
                        )



#import unittest
#import random
#from dynts.utils.populate import populate
#from dynts.utils.populate import datepopulate
#
#__all__ = ['DevTestTS',
#           ]

#class CommonTSUtils(unittest.TestCase):
#
#    def getdata(self, size = 100, delta = 1):
#        date = datepopulate(size = size, delta = delta)
#        data = populate(size = size)
#        return date,data
#
#    def getts(self, returndata = False, delta = 1):
#        date,data = self.getdata(100,delta)
#
#        ts   = TimeSeries(name = 'test', ts = zip(date, data))
#        if returndata:
#            return ts,list(date),list(data)
#        else:
#            return ts
#
#    def assertAlmostEqual(self, exp, recv, tol = 2, msg = None):
#        if msg is None:
#            msg = "Values do not match expected %s, received %s" %(exp, recv)
#        unittest.TestCase.assertAlmostEqual(self, exp,recv, tol, msg)
#
#class DevTestTS(CommonTSUtils):
#
#    def xtestAddTS(self):
#        ts1, _, data1 = self.getts(returndata = True)
#        ts2, _, data2 = self.getts(returndata = True)
#        ts3 = ts2 + ts1
#
#        exp = sum(data1) + sum(data2)
#        recv = sum(ts3.values())
#        self.assertAlmostEqual(exp, recv)
#
#    def xtestAddScalar(self):
#        delta = random.uniform(1.0,100.0)
#        ts, dates, data = self.getts(returndata = True)
#        new_ts = ts + delta
#        for dt, val in zip(dates, data):
#            exp = val + delta
#            recv = new_ts[dt]
#            self.assertAlmostEqual(exp, recv)


################==============

#
# NOT USED DIRECTLY - THIS MODULE IS IMPORTED BY dynts.tests
#


import unittest
from itertools import izip

import dynts
from dynts import dsl

class TestDsl(unittest.TestCase):
    names = ['eur','1eur','eu3r','eur4567']
    
    def testName(self):
        for name in self.names:
            res = dynts.parse(name)
            self.assertTrue(isinstance(res,dsl.Symbol))
            self.assertEqual(name.upper(),str(res))
        
    def testBinaryOperation(self):
        res = dynts.parse('2*GOOG')
        self.assertEqual(len(res.children),2)
        names = res.symbols()
        self.assertEqual(len(names),1)
        self.assertEqual(names[0],str(res.right))
        #ts = dynts.evaluate(res)
        #self.assertEqual(ts.count(),1)
    
    def testAdditionOperation(self):
        res = dynts.parse('YHOO+GOOG')
        self.assertEqual(len(res.children),2)
        names = res.symbols()
        self.assertEqual(len(names),2)
        self.assertEqual(names[0],str(res.left))
        self.assertEqual(names[1],str(res.right))
        
    def testTwoTimeSeries(self):
        '''Get a timeseries and a function and check for consistency'''
        res = dynts.parse('gs:yahoo,min(gs:yahoo,window=30)')
        self.assertEqual(res.symbols(),['GS:YAHOO'])
        self.assertEqual(len(res),2)
        self.assertEqual(str(res[0]),'GS:YAHOO')
        self.assertEqual(str(res[1]),'min(GS:YAHOO,window=30)')
        
    def testBinOpSerieFunction(self):
        '''Get a timeseries and a function and check for consistency'''
        res = dynts.parse('goog:yahoo-ma(goog:yahoo,window=30)')
        self.assertEqual(res.symbols(),['GOOG:YAHOO'])
        self.assertEqual(len(res),2)
        self.assertEqual(str(res[0]),'GOOG:YAHOO')
        self.assertEqual(str(res[1]),'ma(GOOG:YAHOO,window=30)')
        result = dynts.evaluate(res)
        self.assertEqual(len(result.data),1)
        data = result.ts()
        self.assertTrue(dynts.istimeseries(data))
        self.assertEqual(data.count(),1)
        
    def testDataProvider(self):
        result = dynts.evaluate('2*GOOG,GOOG')
        self.assertEqual(len(result.data),1)
        self.assertEqual(result.expression,dynts.parse('2*GOOG,GOOG'))
        data = result.ts()
        self.assertTrue(dynts.istimeseries(data))
        self.assertEqual(data.count(),2)
        ts1 = data.serie(0)
        ts2 = data.serie(1)
        for v1,v2 in izip(ts1,ts2):
            self.assertAlmostEqual(v1,2.*v2)
        
    def testTSName(self):
        '''
        The dslresult should include an attribute 'name' 
        which is the equivalent to the expression passed.
        In situations where multiple timeseries are returned 
        the name should be the concatenation of all the names 
        joined by "__".
        '''
        expressions = ['GOOG+YHOO',
                       '2*GOOG',
                       'GOOG,YHOO',
                       ]
        for expr in expressions:
            result = dynts.evaluate(expr)
            ts = result.ts()
            name = ts.name
            
            expected_name = '__'.join(expr.split(','))
            self.assertEqual(name, expected_name)
            
        

        



from dynts.dsl.ast.astbase import *

class BinMathOp(BinOp):
    
    def __init__(self,left,right,op,op_name):
        self.op_name = op_name
        super(BinMathOp,self).__init__(left,right,op)
    
    def _unwind(self, values, backend, full = False, **kwargs):
        le = self.left.unwind(values, backend, **kwargs)
        ri = self.right.unwind(values, backend, **kwargs)
        return ts_bin_op(self.op_name,le,ri,name=str(self))
    
    def lineardecomp(self):
        if isinstance(self.left,Number):
            if isinstance(self.right,Number):
                raise ValueError
            nv = self.left.value
            d1 = self.right.lineardecomp()
            dr = {}
            for k,v in d1.items():
                dr[k] = nv*v
            return dr
    

class PlusOp(BinMathOp):
    
    def __init__(self,left,right):
        BinMathOp.__init__(self,left,right,'+','add')
    
    def lineardecomp(self):
        ls = self.left.lineardecomp()
        rs = self.right.lineardecomp()
        if ls and rs:
            lc = linearDecomp()
            lc.expand(ls).expand(rs)
            return lc
        else:
            return None
    
    
class MinusOp(BinMathOp):
    def __init__(self,left,right):
        BinMathOp.__init__(self,left,right,'-','sub')
    
    def lineardecomp(self):
        ls = self.left.lineardecomp()
        rs = self.right.lineardecomp()
        if ls and rs:
            lc = linearDecomp()
            lc.expand(ls).expand(rs.changesign())
            return lc
        else:
            return None
        
        
class MultiplyOp(BinMathOp):
    
    def __init__(self,left,right):
        BinMathOp.__init__(self,left,right,'*','mul')
    
    def lineardecomp(self):
        return linearDecomp().append(self)
    
    
class DivideOp(BinMathOp):
    
    def __init__(self,left,right):
        BinMathOp.__init__(self,left,right,'/','div')
    
    def lineardecomp(self):
        return linearDecomp().append(self)

